#!/usr/bin/env python
# encoding: utf-8
"""
MPET_filtering_pipeline.py

Created by Rod Docking on 2011-02-08.
Copyright (c) 2011 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import subprocess
from optparse import OptionParser
from ruffus import *
from time import sleep
from glob import glob

from lib.cluster_job import Cluster_job

PYTHON_EXECUTABLE  = sys.executable
LAUNDRY_SCRIPT     = os.path.dirname(sys.argv[0]) + '/lib/fasta_laundry.py'
INTERPOLATE_SCRIPT = os.path.dirname(sys.argv[0]) + '/lib/interpolate_missing_bwasw_records.py'
FILTER_SCRIPT      = os.path.dirname(sys.argv[0]) + '/lib/filter_MPET_from_bwasw.py'

##################################################
#
#   Option parsing and global names
#
##################################################

#
# Parse command-line options - note that these are GLOBAL
#

def parse_options():
    """Parse options provided on the command line"""
    
    usage = "usage: %prog [options] reference.fa OUTPUT_ROOT READS_LIST"
    parser = OptionParser(usage)
    
    parser.add_option("-c", "--cluster",
        action="store_true", dest="cluster", 
        help="Run jobs on the genesis cluster")
        
    (options, args) = parser.parse_args()
    
    if not (len(args) == 3):
        parser.error("incorrect number of arguments")
                
    return options, args

opt, args = parse_options()

#
# Set up names for output files based on original reference file and input reads files
#

# Names for reference files
original_ref = args[0]
cleaned_ref = args[1] + "-10.fa"
cleaned_ref_faidx = cleaned_ref + ".fai"
cleaned_ref_bwa_index = cleaned_ref + ".rsa"

# Names for alignment files
read_files, sam_files, filtered_sam_files = [], [], []
reads_list = args[2]
output_check_file = reads_list + "_mpet.1.fq"

# Go through the list of read files
for line in open(reads_list, 'r'):
    c = line.split()
    read_files.extend(c)
     
# For now, we're assuming that all read files are paired and in separate files
assert(len(read_files) % 2 == 0)

# Generate intermediate file path names
for read_file in read_files:
     
    # But intermediate files will be in the current directory
    pathed_read_root = os.path.join(os.getcwd(), os.path.basename(read_file))
     
    # Add the name of the SAM output file and filtered SAM output file
    sam_files.append( pathed_read_root + ".sam")
    filtered_sam_files.append( pathed_read_root + ".filtered.sam" )

# Redirect standard error for all subprocess jobs
std_error_log = open('{0}.err'.format(reads_list),'w')

############################################################
#
#   Pipeline functions handled by Ruffus
#
############################################################

#
#   Clean the starting fasta file
#

@files(original_ref, cleaned_ref)
def clean_fasta(input_file, output_file):
    """Strip comments and add line breaks to a fasta file"""
    
    print "{0} {1} -N -l80 {2} {3}".format(PYTHON_EXECUTABLE, LAUNDRY_SCRIPT, input_file, output_file)

    proc = subprocess.Popen(
        [PYTHON_EXECUTABLE, LAUNDRY_SCRIPT, 
          '-N', '-l', '80', input_file, output_file], 
          shell=False, cwd=os.getcwd(), stderr=std_error_log)
    proc.wait()
    
    return

#
#   Index the cleaned fasta file
#

@follows(clean_fasta)
@files(cleaned_ref, cleaned_ref_faidx)
def samtools_faidx(reference, index_file):
    """Run samtools faidx"""
    
    print "samtools faidx {0}".format(reference)
    
    proc = subprocess.Popen(
        ['samtools', 'faidx', reference],
        shell=False, cwd=os.getcwd(), stderr=std_error_log)
    proc.wait()    
    
    return

#
#   Index the cleaned fasta file with bwa index
#

@follows(samtools_faidx)
@files(cleaned_ref, cleaned_ref_bwa_index)
def bwa_index(input_file, output_file):
    """Run bwa index"""
    
    print "bwa index {0}".format(input_file)
    
    proc = subprocess.Popen(
        ['bwa', 'index', input_file], shell=False, cwd=os.getcwd(), stderr=std_error_log)    
    proc.wait()

    return

#
#   Run bwa bwasw to align the reads to the assembly contigs
#

bwasw_parameters = [ [x, y, cleaned_ref] for x, y in zip(read_files, sam_files) ]
@follows(bwa_index)
@files(bwasw_parameters)
def bwa_bwasw(read_file, sam_file, cleaned_ref):
    """Run bwa bwasw"""
    
    print "abyss-tofastq {0} | bwa bwasw -t12 {0} - > {2}".format(
        read_file, cleaned_ref, sam_file)
    
    if opt.cluster:
    
        cmd = "abyss-tofastq {0} | bwa bwasw -t12 {1} - > {2}".format(
            read_file, cleaned_ref, sam_file)
            
        jobname = "bwasw_{0}.sh".format(os.path.basename(read_file))
        
        C = Cluster_job(jobname = jobname, pe = 'ncpus', cpus = '12', host = 'login1')
        C.append(cmd)
        C.submit()

    else:
        
        print "TO-DO: Update this to be like the above (with a pipe from abyss-tofastq)"
        sys.exit()
        # proc = subprocess.Popen(
        #     ['bwa', 'bwasw', '-t4', cleaned_ref, read_file], 
        #     shell=False, cwd=os.getcwd(), stderr=std_error_log,
        #     stdout=open(sam_file, 'w'))    
        # proc.wait()

interpolate_parameters = [ [x , y, z] for x, y, z in zip(sam_files, filtered_sam_files, read_files) ]
@follows(bwa_bwasw)
@files(interpolate_parameters)
def interpolate_sam(sam_file, filtered_sam, read_file):
    """Interpolate missing SAM records, and filter duplicates"""
    
    print "{0} {1} {2} {3} {4}".format(
        PYTHON_EXECUTABLE, INTERPOLATE_SCRIPT, sam_file, read_file, filtered_sam)
        
    proc = subprocess.Popen(
        [PYTHON_EXECUTABLE, INTERPOLATE_SCRIPT, sam_file, read_file, filtered_sam], 
        shell=False, cwd=os.getcwd(), stderr=std_error_log)
    proc.wait()

@follows(interpolate_sam)
@merge(filtered_sam_files, output_check_file, reads_list)
def filter_MPET(filtered_sam_files, output_check_file, reads_list):
    """Generate filtered read files"""
    
    print "{0} {1} {2} {3} {4}".format(
        PYTHON_EXECUTABLE, FILTER_SCRIPT, filtered_sam_files[0], filtered_sam_files[1], reads_list)

    proc = subprocess.Popen(
        [PYTHON_EXECUTABLE, FILTER_SCRIPT, filtered_sam_files[0], filtered_sam_files[1], reads_list], 
        shell=False, cwd=os.getcwd(), stderr=std_error_log)
    proc.wait()
        
    
def main():
    
    pipeline_printout(sys.stdout, [bwa_bwasw], verbose = 4)
    # pipeline_printout_graph(
    #     open('polishing_pipeline_graph.svg','w'), "svg", [polish_assembly])
    pipeline_run([bwa_bwasw], verbose = 4, multiprocess = 1)


if __name__ == '__main__':
    main()

