#!/usr/bin/env python
# encoding: utf-8
"""
SAM_record.py

Created by Rod Docking on 2010-10-08.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import string
import re
import subprocess

cigar_pattern = re.compile('\d+\D')

def bam_open(bam_file):
    """ Return SAM lines from a BAM file """

    proc = subprocess.Popen(
            ['samtools', 'view', bam_file], 
            shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)
            
    return proc.stdout        

def reverse_complement(seq):
    """Return the reverse complement"""

    return seq.translate(string.maketrans("ATCG", "TAGC"))[::-1]

class SAM_record:
    """A simple class for holding a single SAM record"""
    
    def __init__(self, line):
        
        # Raw line
        self.raw_sam = line
        c = line.split('\t')
        
        # Check for header lines
        if c[0] in ["@HD", "@SQ", "@RG", "@PG", "@CO"]:    
            self.header = True
            self.header_list = c
            return
        else:
            self.header = False
                
        # Remaining SAM fields    
        self.qname = c[0]
        self.flag = int(c[1])
        self.rname = c[2]
        self.pos = int(c[3])
        self.mapq = int(c[4])
        self.cigar = c[5]
        self.mrnm = c[6]
        self.mpos = int(c[7])
        self.isize = int(c[8])
        self.seq = c[9]
        self.qual = c[10]
        self.opt_fields = c[11:]

        # Attempt to infer if the reads are numbered (e.g. read/1 read/2)
        #  Otherwise, try to get the read number from the flag
        #  NOT going to try to modify qnames - seems to lead to unexpected behaviour
        # Let scripts try to infer this themselves if they need it...
        # if "/" in c[0][-2:]:
        #     self.qname = c[0][:-2]
        #     self.read = int(c[0][-1:])
        # else:
        #     self.qname = c[0]
        #     if self.first_read():
        #         self.read = 1 
        #     else:
        #         self.read = 2
        

        # Convenience fields
        self.read_length = len(c[9])
        self.strand = "Reverse" if (self.flag & 0x10) else "Forward"  
        if self.first_read():
            self.read = 1
        elif self.second_read():
            self.read = 2
        else:
            self.read = None        
        
        # Store a few of the more common optional fields (some of these may be BWA-specific)
        self.edit_distance = None
        for field in self.opt_fields:
            if "NM" in field:
                self.edit_distance = int(field.split(":")[-1])
        
    def __repr__(self):
        return "<SAM_record: {0}>".format(self.qname)
        
    def __str__(self):
        return """qname (read):\t{0} {1} \
                \nflag:\t\t{2} \
                \nrname (pos):\t{3} {4} \
                \nmapq:\t\t{5} \
                \ncigar:\t\t{6} \
                \nmrnm (pos):\t{7} {8} \
                \nisize:\t\t{9} \
                \nopt_fields:\t{10}\n""".format(
                    self.qname, self.read, self.flag, self.rname, self.pos,
                    self.mapq, self.cigar, self.mrnm, self.mpos, self.isize, 
                    self.opt_fields)    

    # Flag checks
    def paired_in_sequencing(self):
        return True if (self.flag & 0x1) else False
    def proper_pair(self):
        return True if (self.flag & 0x2) else False
    def q_unmapped(self):
        return True if (self.flag & 0x4) else False        
    def m_unmapped(self):
        return True if (self.flag & 0x4) else False        
    def reverse_strand(self):
        return True if (self.flag & 0x10) else False
    def mate_reverse_strand(self):
        return True if (self.flag & 0x20) else False
    def first_read(self):
        return True if (self.flag & 0x40) else False
    def second_read(self):
        return True if (self.flag & 0x80) else False
    def not_primary(self):
        return True if (self.flag & 0x100) else False
    def platform_qc_fail(self):
        return True if (self.flag & 0x200) else False
    def pcr_duplicate(self):
        return True if (self.flag & 0x400) else False
        
    # Utility functions
    def print_fastq(self, re_reverse = True):
        """Print a fastq record from a SAM line"""

        if self.reverse_strand() and re_reverse:
            tmp_seq = reverse_complement(self.seq)
            tmp_qual = self.qual[::-1]
        else:
            tmp_seq = self.seq
            tmp_qual = self.qual

        return "@{0}/{1}\n{2}\n+\n{3}\n".format(
            self.qname, self.read, tmp_seq, tmp_qual)    

    def print_sam(self):
        """Dump the record as SAM"""
        
        fields = [self.qname, self.flag, self.rname, self.pos, self.mapq, 
            self.cigar, self.mrnm, self.mpos, self.isize, self.seq, self.qual]
        fields.extend(self.opt_fields)    
        str_fields = [str(c) for c in fields]
        
        return "{0}".format("\t".join(str_fields))
    
    def cigar_list(self):
        """Split the cigar field into a list"""
        
        return re.findall(cigar_pattern, self.cigar)
    
    def qual_list(self, conversion = 33):
        """Split the quality string into a list of qualities"""    
    
        return [ (ord(qual)-conversion) for qual in self.qual ]
    
    def perfect_match(self, edit_distance_threshold = 0):
        """ Return true if the record is an exact match to it's reference """
        
        # Filter out unmapped reads
        if self.q_unmapped():
            return False
            
        # Confirm that CIGAR is read-length + 'M'
        if not self.cigar == (str(self.read_length) + 'M'):
            return False    
        
        if self.edit_distance > edit_distance_threshold:
            return False

        # If we pass all these checks, return True
        return True
    
if __name__ == '__main__':
    print "Hey, it's SAM_record.py!"