#!/usr/bin/env python
# encoding: utf-8
"""
blat_parsing.py

Created by Rod Docking on 2010-06-18.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys, os

def cat_blat(file_name):
    """Open a file and print each line"""
    
    handle = open(file_name, 'r')
    for line in handle:
        print line,
    handle.close()
    return    

def check_blat_hit(hit, slice_coords, outgoing = True):
    """Check whether a blat hit is consistent with being an incoming or outgoing link
    
        Returns either a number (which is the point on the query where it leaves the target
        and enters the gap, or False)
    
    """

    hit_is_consistent = False

    # For outgoing hits:
    if outgoing:

        # Outgoing, + strand
        if (hit['strand'] == '+'):
            
            # Query:            0 - ---------------- - t_end = t_size
            # Target:   ----------------
            if (hit['q_start'] == 0) and (hit['t_end'] == hit['t_size']):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[0] = hit['q_end']

            # Query:   -------------------------------------- 
            # Target:       t_start = 0 -----------------   t_end = t_size
            elif (hit['t_start'] == 0) and (hit['t_end'] == hit['t_size']):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[0] = hit['q_end']
            
            

        # Outgoing, - strand
        else:
            
            # Query:  0 - ---------------- - q_end = q_size
            # Target:           --------------------- t_end = t_size
            if (hit['q_end'] == hit['q_size']) and (hit['t_end'] == hit['t_size']):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[0] = hit['q_start']
            # Query:  0 - ------------------------------- 
            # Target:           --------------------- t_end = t_size
            elif (hit['t_start'] == 0) and (hit['t_end'] == hit['t_size']):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[0] = hit['q_start']
                
    else:
        
        # Incoming, + strand
        if (hit['strand'] == '+'):
            
            # Query:    -------------- - q_end = q_size
            # Target:           0 ----------------
            if (hit['q_end'] == hit['q_size']) and (hit['t_start'] == 0):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[1] = hit['q_start']
            
            # Query:    ----------------------------- 
            # Target:           0 ---------------- - t_end = t_size
            elif (hit['t_end'] == hit['t_size']) and (hit['t_start'] == 0):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[1] = hit['q_start']
                        
        # Incoming, - strand
        else:

            # Query:            0 - ---------------- - t_end = t_size
            # Target:   ----------------
            if (hit['q_start'] == 0) and (hit['t_end'] == hit['t_size']):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[1] = hit['q_end']

            # Query:   -------------------------------------- 
            # Target:       t_start = 0 -----------------   t_end = t_size
            elif (hit['t_start'] == 0) and (hit['t_end'] == hit['t_size']):
                print "\t\t==> consistent hit"
                hit_is_consistent = True
                slice_coords[0] = hit['q_end']
            
            
            
    return hit_is_consistent, slice_coords    

def estimate_gap_or_overlap_distance(hit_list):
    """From a set of paired blat hits, estimate the positive or negative distance"""

    # Note: when we get here, we've already inferred that we only have a pair of hits to deal with
    # Don't need to worry about strandedness here: we're concerned only with the coordinates against the query

    if not len(hit_list) == 2:
        print "More than two hits to estimate a distance for!"
        sys.exit()

    first_start, first_stop = hit_list[0]['q_start'], hit_list[0]['q_end']
    second_start, second_stop = hit_list[1]['q_start'], hit_list[1]['q_end']     

    if first_start <= second_start:
        distance = second_start - first_stop
    else:
        distance = first_start - second_stop    

    return distance

def filter_hit_list_for_single_hit_per_query(hit_list):
    """Read a hit dictionary constructed by read_blat_file, return a filtered dictionary
    containing only the single longest hit for each query"""
    
    best_hit_dict, new_list = {}, []
    
    for hit in hit_list:
        
        if (hit['q_name'] in best_hit_dict):
            previous_hit = best_hit_dict[hit['q_name']]
            if (hit['match_length'] > previous_hit['match_length']):
                #print "Duplicate hit longer than best hit so far", hit['match_length'], previous_hit['match_length']
                best_hit_dict[hit['q_name']] = hit    
            else:
                pass
                #print "Duplicate hit shorter than best hit so far", hit['match_length'], previous_hit['match_length']
        else:
            best_hit_dict[hit['q_name']] = hit    
    
    for hit in best_hit_dict.keys():
        new_list.append(best_hit_dict[hit])
        
    return new_list

def read_blat_file(blat_file, verbosity = 'Quiet'):
    """Parse a psl-format file, return a list of blat hits"""
    
    # Check that the file exists
    if not os.path.exists(blat_file):
        print "\tCouldn't find alignment file"
        return None
    
    if not verbosity in ['Quiet', 'Raw', 'Full']:
        print "verbosity should be one of 'Quiet', 'Raw', or 'Full' "
        return None
    
    # Open the file for reading
    blat_handle = open(blat_file, 'r')
    hit_list = []
    
    # Check for header
    past_header = True
    
    # For each hit in the file, construct a dictionary and add it to the list
    for line in blat_handle:
        
        if 'psLayout' in line:
            past_header = False
            continue
        elif '-----' in line:
            past_header = True     
            continue
        elif past_header:    
            hit = parse_blat_line(line)
            hit_list.append(hit)    
    
            if verbosity == 'Full':
                print "\n\t{0}bp hit from query {1} ({2}bp) to target {3} ({4}bp)".format(
                    hit['match_length'], hit['q_name'], hit['q_size'], hit['t_name'], hit['t_size'])
                print "\t{0}-{1} on query to {2}-{3} on target {4} strand".format(
                    hit['q_start'], hit['q_end'], hit['t_start'], hit['t_end'], hit['strand'])
                print "\t{0} mismatch base(s) in {1} block(s)".format(hit['num_mismatch'], hit['blockcount'])
            elif verbosity == 'Raw':
                print line.rstrip()
                
    blat_handle.close()    
    
    return hit_list

def parse_blat_line(blat_line):
    """From a line of psl-formatted output, return a dictionary of blat variables"""
    
    c = blat_line.split()

    if not len(c) == 21:
        print "Missing columns from BLAT output!"
        return None
    
    blat_dict = {
            'match_length'          : int(c[0]),
            'num_mismatch'          : int(c[1]),
            'num_rep_match'         : int(c[2]),
            'num_N'                 : int(c[3]),
            'num_q_inserts'         : int(c[4]),
            'num_q_insert_bases'    : int(c[5]),
            'num_t_inserts'         : int(c[6]),
            'num_t_insert_bases'    : int(c[7]),
            'strand'                : c[8],
            'q_name'                : c[9],
            'q_size'                : int(c[10]),
            'q_start'               : int(c[11]),
            'q_end'                 : int(c[12]),
            't_name'                : c[13],
            't_size'                : int(c[14]),
            't_start'               : int(c[15]),
            't_end'                 : int(c[16]),
            'blockcount'            : int(c[17]),
            'blocksizes'            : c[18],
            'q_starts'              : c[19],
            't_starts'              : c[20],
    }
    
    return blat_dict



def main():
    """This module contains functions having to do with parsing blat output"""
    
    hit_list = read_blat_file(sys.argv[1], verbosity = 'Full')
    for hit in hit_list:
        print hit


if __name__ == '__main__':
    main()

