#!/usr/bin/env python
# encoding: utf-8
"""
common_utils.py

Written by Rod Docking (rdocking@bcgsc.ca)
"""

import re, subprocess, os, string, sys

def build_header_dict_from_fai(ref_file):
    """From a reference file, build the fasta-index file and return a dictionary of sequence names and lengths
       This is intended to replace the parse_header_dict function below - we don't want building the header file to be a requirement
    """
    
    if not os.path.exists(ref_file):
        print "Reference file does not exist!"
        sys.exit()
    
    fai_file = '{0}.fai'.format(ref_file)
    
    if not os.path.exists(fai_file):
        subprocess.call(['samtools', 'faidx', ref_file])
    
    header_dict = {}
    handle_in = open(fai_file, 'r')
    for line in handle_in:
        columns = line.split()
        ref_name, ref_length = columns[0], int(columns[1])
        header_dict[ref_name] = ref_length

    return header_dict
    

def coord_to_seq(coord):
    """Convert a samfile coordinate to just the sequence name: e.g. Seq1:0-100 ==> Seq1"""
    
    if ":" in coord:
        m = re.match("^(.+)(\:)", coord)
        return (m.group(0)).rstrip(":")
    else:
        print "Input doesn't look like a coordinate: {0}".format(coord)
        return coord

def est_to_clone(est_name):
    """Convert an EST-style name to a clone-style name
     (e.g. DPO101.C21_G21 ==> DPO101_G21)""" 
    
    if "." in est_name:
        pattern = re.compile(r'^(\w+).(.+)_(\w+)$')
        groups = pattern.search(est_name).groups()
        clone_name = '{0}_{1}'.format(groups[0], groups[2])
        return clone_name
    else:
        #print "Input doesn't look like an EST name: {0}".format(est_name)
        return est_name

def count_fastq_reads(fq_file):
    """Count the number of reads in a fastq file"""
    
    fq_handle = open(fq_file, 'r')
    count = len(fq_handle.readlines())/4
    fq_handle.close()
    
    return count

def get_current_svn_rev(path='.'):
    """Return the current svn revision of the script"""
    
    script_dir = os.path.dirname(path)
    if script_dir == '':
        script_dir = '.'
    if not os.path.exists("{0}/.svn".format(script_dir)):
        return None
    process = subprocess.Popen(['svn', 'info'], shell=False, cwd = script_dir, stdout=subprocess.PIPE).communicate()[0]
    lines = process.split('\n')
    for l in lines:
        if "Revision" in l: 
            return l.split()[1]

def get_flank_coords(seq_name, seq_length, flanksize):
    """ From a sequence name and length, return the flank coordinates """

    if (seq_length > (flanksize*2)):
        flanks = ("{0}:0-{1}".format(seq_name, flanksize),  
            "{0}:{1}-{2}".format(seq_name, seq_length-flanksize, seq_length))
    else:
        flanks = ("{0}:0-{1}".format(seq_name, seq_length), )

    return flanks      

def get_single_flank(seq_name, seq_length, flanksize, outgoing = True):
    """ From sequence name, length, and flank, return a SAM-style coordinate"""

    seq_name_only = seq_name.rstrip('+-')

    if (seq_length < flanksize):
        flank = "{0}:1-{1}".format(seq_name_only, seq_length)
    elif (outgoing and ('+' in seq_name)) or (not outgoing and ('-') in seq_name):
        flank = "{0}:{1}-{2}".format(seq_name_only, seq_length-flanksize, seq_length)
    elif (not outgoing and ('+') in seq_name) or (outgoing and ('-') in seq_name):
        flank = "{0}:1-{1}".format(seq_name_only, flanksize)
    else:
        print "Failed to extract a flank coordinate"
        sys.exit()
    
    return flank        

def invert_pgv_link(edge_to_invert):
    """Invert a pygraphviz-style link"""

    # For each node name, convert '+' to '-' and vice-versa
    first_node = edge_to_invert[1][:-1] + edge_to_invert[1][-1:].translate(string.maketrans("+-","-+"))
    second_node = edge_to_invert[0][:-1] + edge_to_invert[0][-1:].translate(string.maketrans("+-","-+"))

    inverted_edge = (first_node, second_node)

    return inverted_edge

def invert_node_name(node_to_invert):
    """Invert a node name, swapping plus for minus"""
    
    return node_to_invert[:-1] + node_to_invert[-1:].translate(string.maketrans("+-","-+"))

def summarize_graph(G):
    """Print quick summary statistics for the graph"""

    result_string = "\nGraph summary:\n"
    result_string = result_string + "\tThere are {0} nodes and {1} edges in the graph\n".format(G.number_of_nodes(),  G.number_of_edges())
    result_string = result_string + "\tGraph is directed: {0}\n\tGraph is strict:{1}\n".format(G.is_directed(), G.is_strict()) 

    from_blat, from_bam = 0, 0

    # For each edge, check that it's inverse is present
    all_inverted_edges_present = True
    for edge in G.iteredges():

        if edge.attr['source'] == 'bam':
            from_bam += 1
        elif edge.attr['source'] == 'blat':
            from_blat += 1

        inverted_edge = invert_pgv_link(edge)
        if not G.has_edge(inverted_edge):
            all_inverted_edges_present = False

    result_string = result_string + "\tBLAT-inferred links: {0}\n\tBAM-inferred links: {1}\n".format(from_blat, from_bam)
    result_string = result_string + "\tAll edges are present in both orientations: {0}\n".format(all_inverted_edges_present)

    return result_string
    # 
    # 
    # print "\nGraph summary:"
    # print "\tThere are", G.number_of_nodes(), "nodes and", G.number_of_edges(), "edges in the graph"
    # print "\tGraph is directed: ", G.is_directed()
    # print "\tGraph is strict: ", G.is_strict()
    # 
    # from_blat, from_bam = 0, 0
    # 
    # # For each edge, check that it's inverse is present
    # all_inverted_edges_present = True
    # for edge in G.iteredges():
    # 
    #     if edge.attr['source'] == 'bam':
    #         from_bam += 1
    #     elif edge.attr['source'] == 'blat':
    #         from_blat += 1
    #     
    #     inverted_edge = invert_pgv_link(edge)
    #     if not G.has_edge(inverted_edge):
    #         all_inverted_edges_present = False
    # 
    # print "\tBLAT-inferred links: ", from_blat
    # print "\tBAM-inferred links: ", from_bam
    # 
    # print "\tAll edges are present in both orientations: ", all_inverted_edges_present    
    # print "\n"
    # 
    # return


def parse_header_dict(header_file):
    """From a samtools-style header dictionary, return a dictionary of reference names and lengths

    i.e., from a BAM file:
    samtools view -H round3_align_all.sorted.bam | grep ^@SQ > hsRef.dict    
    """
    
    if not os.path.exists(header_file):
        print "Headers file does not exist!"
        sys.exit()
    
    header_dict = {}
    handle_in = open(header_file, 'r')
    for line in handle_in:
        
        if not '@SQ' in line:
            continue
        
        columns = line.split()
        ref_name, ref_length = columns[1][3:], int(columns[2][3:])
        header_dict[ref_name] = ref_length

    return header_dict


def build_seq_dict(reference, header_dict = None):
    """Parse a fasta file and return a dictionary with keys of ref name, value of sequence"""
    
    fasta_dict = {}
    fasta_in = open(reference,'r')
    current_header, current_seq = None, ''
    for line in fasta_in:
        
        if (">" in line) and current_header:
            fasta_dict[current_header] = current_seq
            current_seq = ''
            current_header = line.split()[0].lstrip('>')
        elif (">" in line) and (not current_header):
            current_header = line.split()[0].lstrip('>')        
        else:    
            current_seq = current_seq + line.rstrip()
    
    fasta_in.close()
    fasta_dict[current_header] = current_seq
    
    #print "Finished with ", len(fasta_dict.keys()), " fasta sequences found"
    
    #If a header dictionary is supplied, check that they agree
    if header_dict:
        for key in fasta_dict.keys():
        
            #running_total += len(fasta_dict[key])
            #print key, len(fasta_dict[key]), "Running total: ", running_total
        
            if not (len(fasta_dict[key]) == header_dict[key]):
                print "Discrepancy between header dictionary and reference file!"
                sys.exit()
    
    return fasta_dict

def build_seq_dict_with_header_comments(reference):
    """Parse a fasta file and return a dictionary with keys of ref name, value of [header_comment, sequence]"""

    fasta_dict = {}
    fasta_in = open(reference,'r')
    current_header, current_seq = None, ''
    for line in fasta_in:

        if (">" in line) and current_header:
            fasta_dict[current_header] = [current_header_comment, current_seq]
            current_seq = ''
            current_header = line.split()[0].lstrip('>')
            current_header_comment_list = line.split()[1:]
            current_header_comment = " ".join(current_header_comment_list)
        elif (">" in line) and (not current_header):
            current_header = line.split()[0].lstrip('>')
            current_header_comment_list = line.split()[1:]        
            current_header_comment = " ".join(current_header_comment_list)
        else:    
            current_seq = current_seq + line.rstrip()

    fasta_in.close()
    fasta_dict[current_header] = [current_header_comment, current_seq]

    print "Finished with ", len(fasta_dict.keys()), " fasta sequences found"
    return fasta_dict

def build_seq_length_dict(reference):
    """ Parse a fasta file and return a dictionary with keys of ref name, values of ref length"""

    # OLD VERSION - build directly from fasta file
    # should migrate code to the parse_header_dict function above
    
    if not os.path.exists(reference):
        return [], []
        
    print "Parsing assembly file for ref names and lengths: ",
    fasta_in = open(reference, 'r')
    current_seq, current_header = '', ''
    ref_dict = {}

    for line in fasta_in.readlines():

        if (current_seq != '') and (">" in line):
            ref_length = get_seq_length(current_seq)
            ref_dict[current_header] = ref_length
            current_header, current_seq = '', ''
        if ">" in line:
            current_header = line.split()[0].lstrip('>')
        else:
            current_seq = current_seq + line        

    ref_length = get_seq_length(current_seq)
    ref_dict[current_header] = ref_length

    fasta_in.close()
    print " Done"

    return ref_dict

def read_single_entry_fasta(fasta_file):
    """ Read a single-entry fasta file, returning a single string with the sequence"""
    
    # Check that the file exists
    if not os.path.exists(fasta_file):
        print "\tCouldn't find fasta file"
        return None
        
    # Retrieve the sequence
    seq = ''
    seq_in = open(fasta_file,'r')
    for line in seq_in:
        if not ">" in line:
            seq = seq + line.rstrip()
    seq_in.close()
    
    return seq    
    
def get_seq_length(seq):
    """ Get length of a multi-line fasta entry"""

    seq_lines = seq.split("\n")
    long_seq = ''
    for line in seq_lines:
        long_seq = long_seq + line.rstrip()

    return len(long_seq)

def reverse_complement(seq):
    """Return the reverse complement"""

    return seq.translate(string.maketrans("ATCGatcgNn", "TAGCtagcNn"))[::-1]

def invert_link(edge_to_invert):
    """From an ABySS-style edge name, return the inverse"""

    split = edge_to_invert.split()
     
    first_node = split[2][:-1] + split[2][-1:].translate(string.maketrans("+-","-+"))
    second_node = split[0][:-1] + split[0][-1:].translate(string.maketrans("+-","-+"))
          
    return first_node + " -> " + second_node

def read_to_min_kmers(read, kmer_size):
    """From a single read sequence and kmer size, return a list of the fewest kmers to cover the whole read"""

    # For now, needs to handle reads of 75 being split into first and last
    # Should be general though
    
    if (len(read) < kmer_size):
        print "Error! Read is shorter than kmer length!"
        sys.exit()

    if (len(read) % kmer_size) == 0:
        num_kmers = (len(read) / kmer_size)
    else:    
        num_kmers = (len(read) / kmer_size) + 1
    
    kmer_list = []
    kmer_list.append(read[:kmer_size])
    
    if num_kmers > 2:
        for i in range(num_kmers-2):
            offset = (i * kmer_size) + 1
            kmer = read[offset:(offset+kmer_size)]
            kmer_list.append(kmer)

    kmer_list.append(read[-kmer_size:])

    return kmer_list
    
def read_to_kmers(read, kmer_size):
    """From a single read sequence and kmer size, return a list of kmers"""

    if (len(read) < kmer_size):
        print "Error! Read is shorter than kmer length!"
        sys.exit()
        
    num_kmers = len(read) - kmer_size + 1
    kmer_list = []
    
    for i in range(num_kmers):
        kmer = read[i:(i+kmer_size)]
        kmer_list.append(kmer)
    
    return kmer_list

def retrieve_reads_from_bam(bamfile, coord = ''):
    """From a BAM file name and coordinate, return a list of reads"""
    
    read_list = []
    process = subprocess.Popen(['samtools', 'view', bamfile, coord],
        shell=False, cwd=os.getcwd(), stdout=subprocess.PIPE)    
    handle = process.stdout
    
    for line in handle:
        read_list.append(line.rstrip())
    
    handle.close()
    
    return read_list    
     
def parse_sam_flag(flag):
    """Parse the SAM flag field"""
    
    result = ''
    if flag & 0x1:
        result = result + '\nread paired in sequencing'
    if not flag & 0x1:
        result = result + '\nread NOT paired in sequencing'
    
    if flag & 0x2:
        result = result + '\nread mapped in proper pair'
    if not flag & 0x2:
        result = result + '\nread NOT mapped in proper pair'
    
    if flag & 0x4:
        result = result + '\nquery sequence is unmapped'
    if not flag & 0x4:
        result = result + '\nquery sequence is mapped'
        
        if flag & 0x10:
            result = result + '\nquery on reverse strand'
        if not flag & 0x10:
            result = result + '\nquery on forward strand'        
    
    if flag & 0x8:
        result = result + '\nmate is unmapped'
    if not flag & 0x8:
        result = result + '\nmate is mapped'
        
        if flag & 0x20:
            result = result + '\nmate on reverse strand'
        if not flag & 0x20:
            result = result + '\nmate on forward strand'
    
    if flag & 0x40:
        result = result + '\nread1'
    if flag & 0x80:
        result = result + '\nread2'
    if flag & 0x100:
        result = result + '\nalignment is not primary'
    if flag & 0x200:
        result = result + '\nread fails platform/vendor checks'
    if flag & 0x400:
        result = result + '\nPCR duplicate/optical duplicate'
    
    return result
    
def long_seq_to_lines_of_fifty(long_seq, line_length = 50):
    """From a long string, split into lines of 50 characters"""
    
    num_lines = len(long_seq) / line_length
    last_line = len(long_seq) - (num_lines * line_length)
            
    split_list = []
    for i in range(num_lines):
        trim_line = long_seq[(i * line_length):(i * line_length) + line_length]
        split_list.append(trim_line)
    
    if not last_line == 0:
        trim_line = long_seq[-last_line:]
        split_list.append(trim_line)

    seq_to_return = "\n".join(split_list) + '\n'

    # Old and slow version!
    # seq_to_return = ''
    # 
    # while(len(long_seq) > line_length):
    #     seq_to_return = seq_to_return + long_seq[:line_length] + "\n"
    #     long_seq = long_seq[line_length:]  
    # seq_to_return = seq_to_return + long_seq.rstrip() + '\n'
    
    return seq_to_return    