#!/usr/bin/env python
# encoding: utf-8
"""
filter_MPET_from_bwasw.py

Created by Rod Docking on 2010-10-07.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre.
All rights reserved.
"""

import sys
from SAM_record import SAM_record, reverse_complement

def infer_orientation_and_isize(sr_one, sr_two):
    """From two SAM_records, infer the pair orientation and insert size"""
    
    # Note that this only should apply to reads aligned to the same contig
    
    # RR
    if ( sr_one.reverse_strand() and sr_two.reverse_strand() ):
        return 'RR', 0
    
    # FF
    elif (not sr_one.reverse_strand() ) and (not sr_two.reverse_strand()):
        return 'FF', 0
    
    # Read one forward, read two reverse
    elif ( (not sr_one.reverse_strand()) and sr_two.reverse_strand() ):
        
        if (sr_one.pos < sr_two.pos):
            return 'FR', sr_two.pos + sr_two.read_length - sr_one.pos
        else:
            return 'RF', sr_two.pos + sr_two.read_length - sr_one.pos
    
    # Read one reverse, read two forward
    elif ( sr_one.reverse_strand() and (not sr_two.reverse_strand()) ):
        
        if (sr_one.pos < sr_two.pos):
            return 'RF', sr_two.pos + sr_two.read_length - sr_one.pos
        else:
            return 'FR', sr_one.pos + sr_one.read_length - sr_two.pos
    
    else:
        print "Some possibilities have been missed!!"
        print sr_one
        print sr_two
        sys.exit()

def split_read(rec):
    """From a clipped read, return the coordinates of PET-like and MPET-like fragments"""
    
    cigar_list = rec.cigar_list()
    
    if len(cigar_list) != 2:
        #sys.stderr.write("Strange cigar, skipping: {0}\n".format(rec.cigar))
        return '', '', '', ''
    
    # For a record on the reverse strand, the cigar list needs to be reversed
    #   As well as the sequence and quality being rc'ed and reversed
    if rec.reverse_strand():
        cigar_list.reverse()
        rec.seq = reverse_complement(rec.seq)
        rec.qual = rec.qual[::-1]
    
    block_one_size = int(cigar_list[0][:-1])
    
    mpet_seq = rec.seq[:block_one_size]
    mpet_qual = rec.qual[:block_one_size]
    pet_seq = rec.seq[block_one_size:]
    pet_qual = rec.qual[block_one_size:]
    
    return pet_seq, pet_qual, mpet_seq, mpet_qual

# def debug_pair(sr_one, sr_two, headers = None):
#     """Print summary information for a pair of reads to aid in debugging"""
#     
#     # Print summary and pause
#     print sr_one.brief()
#     print sr_two.brief()
#     
#     same_contig = True if (sr_one.rname == sr_two.rname) else False
#     if same_contig:
#         orientation, isize = infer_orientation_and_isize(sr_one, sr_two)
#         print orientation, isize
#     else:
#         print "Different contigs!"
#     
#     full_match = (str(sr_one.read_length) + 'M')
#     if sr_one.cigar == full_match:
#         clean_read = sr_one
#         clip_read = sr_two
#     else:
#         clean_read = sr_two
#         clip_read = sr_one
#     
#     print "Clean: ", clean_read.read
#     print "Clipped: ", clip_read.read
#     
#     pet_seq, pet_qual, mpet_seq, mpet_qual = split_read(clip_read)
#     print clip_read.seq
#     print mpet_seq
#     print pet_seq
#     
#     if headers:
#         print "Contig lengths:"
#         print sr_one.rname, '==>', headers[sr_one.rname]
#         print sr_two.rname, '==>', headers[sr_two.rname]
#         
#         for sr in (sr_one, sr_two):
#             
#             if sr.strand == 'Forward':
#                 sr.min_FR = headers[sr.rname] - sr.pos
#                 sr.min_RF = sr.pos + sr.read_length
#             else:
#                 sr.min_FR = sr.pos + sr.read_length
#                 sr.min_RF = headers[sr.rname] - sr.pos
#         
#         print "Minimum FR distance: {0} + {1} = {2}".format(
#                 sr_one.min_FR, sr_two.min_FR, sr_one.min_FR + sr_two.min_FR)
#         print "Minimum RF distance: {0} + {1} = {2}".format(
#                 sr_one.min_RF, sr_two.min_RF, sr_one.min_RF + sr_two.min_RF)
#     
#     raw_input()
#     
#     return

def match_percentage(sam_record):
    """Infer the percentage of the read in the matching portion"""
    
    # Find all the 'Match' components of the cigar string and convert to int
    match_part = [int(x.rstrip('M')) for x in sam_record.cigar_list() if 'M' in x]
    
    # Check that there's only a single matching component
    if not len(match_part) == 1:
        return None
    
    # Return the percentage of the total read length in the matching portion
    return float(match_part[0])/float(sam_record.read_length)*100.0

def check_neither_clipped_same_contig(results, sr_one, sr_two, handles):
    """Infer read status for pairs where:
        - Both reads aligned
        - Neither is clipped
        - The reads align to the same contig
    """
    
    orientation, isize = infer_orientation_and_isize(sr_one, sr_two)
    if orientation == 'FR' and abs(isize) < 1000:
        results['neither_clipped_same_contig_FR'] += 1
        handles['p1'].write(sr_one.print_fastq())
        handles['p2'].write(sr_two.print_fastq())
    
    elif orientation == 'RF' and abs(isize) > 800:
        results['neither_clipped_same_contig_RF'] += 1
        handles['mp1'].write(sr_one.print_fastq())
        handles['mp2'].write(sr_two.print_fastq())
    
    else:
        results['neither_clipped_same_contig_ambig'] += 1
    
    return results

def check_neither_clipped_diff_contig(results, sr_one, sr_two, headers, handles):
    """Infer read status for pairs where:
        - Both reads aligned
        - Neither is clipped
        - The reads align to the same contig
    """
        
    # Compute the minimum plausible RF and FR distances
    for sr in (sr_one, sr_two):
        
        if sr.strand == 'Forward':
            sr.min_FR = headers[sr.rname] - sr.pos
            sr.min_RF = sr.pos + sr.read_length
        else:
            sr.min_FR = sr.pos + sr.read_length
            sr.min_RF = headers[sr.rname] - sr.pos
    
    min_FR = sr_one.min_FR + sr_two.min_FR
    min_RF = sr_one.min_RF + sr_two.min_RF
    
    # See if we can confidently assign them to RF or FR
    if min_FR > 800 and min_RF < 6000:
        results['neither_clipped_diff_contig_RF'] += 1
        handles['mp1'].write(sr_one.print_fastq())
        handles['mp2'].write(sr_two.print_fastq())
    
    elif min_FR < 800 and min_RF > 6000:
        results['neither_clipped_diff_contig_FR'] += 1
        handles['p1'].write(sr_one.print_fastq())
        handles['p2'].write(sr_two.print_fastq())
    
    else:
        results['neither_clipped_diff_contig_ambig'] += 1
        #debug_pair(sr_one, sr_two, headers)
    
    return results

def check_one_clipped_same_contig(results, sr_one, sr_two, handles):
    """Infer read status for pairs where:
        - Both reads aligned
        - Neither is clipped
        - The reads align to the same contig
    """
        
    #debug_pair(sr_one, sr_two)
    full_match = (str(sr_one.read_length) + 'M')
    if sr_one.cigar == full_match:
        clean_read = sr_one
        clip_read = sr_two
    else:
        clean_read = sr_two
        clip_read = sr_one
    
    clean_read.read = 1
    clip_read.read = 2
    pet_seq, pet_qual, mpet_seq, mpet_qual = split_read(clip_read)
    
    if (len(pet_seq) >= 20):
        handles['p1'].write(clean_read.print_fastq())
        handles['p2'].write("@{0}/{1}\n{2}\n+\n{3}\n".format(
            clip_read.qname, clip_read.read, pet_seq, pet_qual))
    if (len(mpet_seq) >= 20):
        handles['mp1'].write(clean_read.print_fastq())
        handles['mp2'].write("@{0}/{1}\n{2}\n+\n{3}\n".format(
            clip_read.qname, clip_read.read, mpet_seq, mpet_qual))
    
    return results

def check_one_clipped_diff_contig(results, sr_one, sr_two, handles):
    """Infer read status for pairs where:
        - Both reads aligned
        - Neither is clipped
        - The reads align to the same contig
    """
    
    # NOTE: this is exactly the same as check_one_clipped_same_contig
    #  at the moment.  Need to decide whether we're actually going
    #  to do different things.  If not, one should be deleted.
        
    #debug_pair(sr_one, sr_two, headers)
    full_match = (str(sr_one.read_length) + 'M')
    if sr_one.cigar == full_match:
        clean_read = sr_one
        clip_read = sr_two
    else:
        clean_read = sr_two
        clip_read = sr_one
    
    clean_read.read = 1
    clip_read.read = 2
    pet_seq, pet_qual, mpet_seq, mpet_qual = split_read(clip_read)
    
    if (len(pet_seq) >= 20):
        handles['p1'].write(clean_read.print_fastq())
        handles['p2'].write("@{0}/{1}\n{2}\n+\n{3}\n".format(
            clip_read.qname, clip_read.read, pet_seq, pet_qual))
    if (len(mpet_seq) >= 20):
        handles['mp1'].write(clean_read.print_fastq())
        handles['mp2'].write("@{0}/{1}\n{2}\n+\n{3}\n".format(
            clip_read.qname, clip_read.read, mpet_seq, mpet_qual))
    
    return results

def parse_sam_pair(results, sr_one, sr_two, headers, handles):
    """Parse a pair of SAM records"""
    
    # Going to change this to use a threshold instead
    full_match = (str(sr_one.read_length) + 'M')
    
    # Ignore minor amounts of clipping
    for sr in (sr_one, sr_two):
        if match_percentage(sr) >= 95.0:
            sr.cigar = full_match
    
    # First check whether 0, 1, or 2 reads are aligned:
    if (sr_one.q_unmapped() and sr_two.q_unmapped()):
        results['neither_aligned'] += 1
    elif (sr_one.q_unmapped() or sr_two.q_unmapped()):
        results['one_aligned'] += 1
    else:
        results['both_aligned'] += 1
        
        # Check whether both are aligned to the same contig
        same_contig = True if (sr_one.rname == sr_two.rname) else False
        
        # Check whether 0, 1, or 2 are clipped:
        if (sr_one.cigar == full_match) and (sr_two.cigar == full_match):
            results['neither_clipped'] += 1
            
            # Same contig or different contigs:
            if same_contig:
                results['neither_clipped_same_contig'] += 1
                results = check_neither_clipped_same_contig(
                           results, sr_one, sr_two, handles)
            else:
                results['neither_clipped_diff_contig'] += 1
                results = check_neither_clipped_diff_contig(
                           results, sr_one, sr_two, headers, handles)
        
        # One read is clipped:
        elif (sr_one.cigar == full_match) or (sr_two.cigar == full_match):
            results['one_clipped'] += 1
            
            # Same or different contigs
            if same_contig:
                results['one_clipped_same_contig'] += 1
                results = check_one_clipped_same_contig(
                           results, sr_one, sr_two, handles)
            else:
                results['one_clipped_diff_contig'] += 1
                results = check_one_clipped_diff_contig(
                           results, sr_one, sr_two, handles)
        
        # Both reads are clipped
        else:
            results['both_clipped'] += 1
    
    return results

def main():
    """From paired bwa bwasw sam files, infer mate-pair and clipping information"""
    
    # Allow reads to be in either one or two SAM files
    if len(sys.argv) == 4:
        handle_one = open(sys.argv[1])
        handle_two = open(sys.argv[2])
    elif len(sys.argv) == 3:
        handle_one = open(sys.argv[1])
        handle_two = handle_one
    else:
        print "Usage: {0} reads1.sam [reads2.sam] output_prefix".format(
            sys.argv[0])
        sys.exit()
    
    # Output file handles
    handles = {
                'p1' : open(sys.argv[-1] + "_pet_1.fq", 'w'),
                'p2' : open(sys.argv[-1] + "_pet_2.fq", 'w'),
                'mp1' : open(sys.argv[-1] + "_mpet_1.fq", 'w'),
                'mp2' : open(sys.argv[-1] + "_mpet_2.fq", 'w')
              }
    
    results = {
                    'total'                             : 0,
                    'neither_aligned'                   : 0,
                    'one_aligned'                       : 0,
                    'both_aligned'                      : 0,
                    'neither_clipped'                   : 0,
                    'one_clipped'                       : 0,
                    'both_clipped'                      : 0,
                    'neither_clipped_same_contig'       : 0,
                    'neither_clipped_same_contig_FR'    : 0,
                    'neither_clipped_same_contig_RF'    : 0,
                    'neither_clipped_same_contig_ambig' : 0,
                    'neither_clipped_diff_contig'       : 0,
                    'neither_clipped_diff_contig_FR'    : 0,
                    'neither_clipped_diff_contig_RF'    : 0,
                    'neither_clipped_diff_contig_ambig' : 0,
                    'one_clipped_same_contig'           : 0,
                    'one_clipped_diff_contig'           : 0,
                    'confused'                          : 0
               }
    
    headers = {}
    counter = 0
    
    while True:
        
        # Read a line from both files
        line_one = handle_one.readline()
        if not line_one:
            break
        line_two = handle_two.readline()
        
        # Collect contig lengths, break when we stop receving new lines
        if line_one[0] == '@':
            if "@SQ" in line_one:
                c = line_one.split()
                c_id = c[1].split(':')[-1]
                c_len = int(c[2].split(':')[-1])
                headers[c_id] = c_len
            continue
        
        sr_one = SAM_record(line_one)
        sr_one.read = 1
        sr_two = SAM_record(line_two)
        sr_two.read = 2
        
        # Make sure the files are in sync
        if not (sr_one.qname == sr_two.qname):
            print "Files are out of sync!!"
            print line_one
            print line_two
            break
        
        # Parse the SAM pair
        results['total'] += 1
        results = parse_sam_pair(
            results, sr_one, sr_two, headers, handles)
        
        # Simple counter
        counter += 1
        if counter % 1000000 == 0:
            sys.stderr.write("Parsed {0} pairs...\n".format(counter))
            #break
    
    # Close all file-handles
    handle_one.close()
    handle_two.close()
    [handle.close for handle in handles.values()]
    
    # Get a dictionary of percentages
    percentages = dict([(x, float(results[x])/float(results['total'])*100.0)
                        for x in results.keys()])
    
    # Results
    print "\n(1) Read alignment status:"
    print "Total:              {0:10d}".format(results['total'])
    print "Neither aligned:    {0:10d}\t{1:5.1f}%\t==> DONE".format(
            results['neither_aligned'],
            percentages['neither_aligned'])
    print "One aligned:        {0:10d}\t{1:5.1f}%\t==> DONE".format(
            results['one_aligned'],
            percentages['one_aligned'])
    print "Both aligned:       {0:10d}\t{1:5.1f}%\t==> (2)".format(
            results['both_aligned'],
            percentages['both_aligned'])
    
    print "\n(2) Pairs where both reads aligned:"
    print "Neither clipped:    {0:10d}\t{1:5.1f}%\t==> (3)".format(
            results['neither_clipped'],
            percentages['neither_clipped'])
    print "One clipped:        {0:10d}\t{1:5.1f}%\t==> (4)".format(
            results['one_clipped'],
            percentages['one_clipped'])
    print "Both clipped:       {0:10d}\t{1:5.1f}%\t==> DONE".format(
            results['both_clipped'],
            percentages['both_clipped'])
    
    print "\n(3) Pairs where both reads aligned and NEITHER alignment is clipped:"
    print "Same contig:        {0:10d}\t{1:5.1f}%\t==> (5)".format(
            results['neither_clipped_same_contig'],
            percentages['neither_clipped_same_contig'])
    print "Different contigs:  {0:10d}\t{1:5.1f}%\t==> (6)".format(
            results['neither_clipped_diff_contig'],
            percentages['neither_clipped_diff_contig'])
    
    print "\n(4) Pairs where both reads aligned and ONE alignment is clipped:"
    print "Same contig:        {0:10d}\t{1:5.1f}%\t==> (7)".format(
            results['one_clipped_same_contig'],
            percentages['one_clipped_same_contig'])
    print "Different contigs:  {0:10d}\t{1:5.1f}%\t==> (8)".format(
            results['one_clipped_diff_contig'],
            percentages['one_clipped_diff_contig'])
    
    print "\n(5) Both reads aligned, neither clipped, same contig:"
    print "FR:                 {0:10d}\t{1:5.1f}%".format(
            results['neither_clipped_same_contig_FR'],
            percentages['neither_clipped_same_contig_FR'])
    print "RF:                 {0:10d}\t{1:5.1f}%".format(
            results['neither_clipped_same_contig_RF'],
            percentages['neither_clipped_same_contig_RF'])
    print "Tandem / Ambiguous: {0:10d}\t{1:5.1f}%".format(
            results['neither_clipped_same_contig_ambig'],
            percentages['neither_clipped_same_contig_ambig'])
    
    print "\n(6) Both reads aligned, neither clipped, different contigs:"
    print "FR:                 {0:10d}\t{1:5.1f}%".format(
            results['neither_clipped_diff_contig_FR'],
            percentages['neither_clipped_diff_contig_FR'])
    print "RF:                 {0:10d}\t{1:5.1f}%".format(
            results['neither_clipped_diff_contig_RF'],
            percentages['neither_clipped_diff_contig_RF'])
    print "Tandem / Ambiguous: {0:10d}\t{1:5.1f}%".format(
            results['neither_clipped_diff_contig_ambig'],
            percentages['neither_clipped_diff_contig_ambig'])
    
    print "\n(7) Both reads aligned, one clipped, same contig:"
    
    
    print "\n(8) Both reads aligned, one clipped, different contig:"

if __name__ == '__main__':
    main()

