#!/usr/bin/env python
# encoding: utf-8
"""
pileup_record.py

Created by Rod Docking on 2010-10-18.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys
import os
import re

class PileupError(Exception):
    """Simple exception class for parsing errors"""
    def __init__(self, message):
        self.message = message
        print "Parsing error: {0}".format(message)

class Pileup_record:
    """Class for encapsulating a single samtools-style pileup record"""
    
    def __init__(self, line):
        """Parse a line of samtools-style pileup information"""
        
        c = line.split('\t')

        if not ((len(c) == 10) or (len(c) == 15)):
            raise PileupError, "improper number of columns - expecting 10 or 15, found {0}\n{1}".format(len(c), line)
        
        # Fields that should always be present
        self.raw = line
        self.rname = c[0]
        self.pos = int(c[1])
        self.ref_base = c[2]
        self.con_base = c[3]
        self.con_qual = int(c[4])
        self.snp_qual = int(c[5])
        self.rms_map_qual = int(c[6])
        self.num_reads = int(c[7])
        
        # Fields are different for SNP vs. indel lines
        if len(c) == 10:
            self.type = 'SNP'
            self.read_bases = c[8]
            self.base_qual = c[9]
        elif len(c) == 15:
            self.type = 'indel'
            self.indel_allele_one = c[8]
            self.indel_allele_two = c[9]
            self.indel_allele_one_support = int(c[10])
            self.indel_allele_two_support = int(c[11])
            self.indel_allele_three_support = int(c[12])            
    
    def __repr__(self):
        return "<pileup_record: {0} {1}>".format(self.rname, self.pos)
    
    def __str__(self):
        
        if self.type == 'SNP':
            pileup_string = "SNP: {0} at: {1}\tRef: {2}\tCall: {3}\n".format(self.rname, self.pos, self.ref_base, self.con_base)
            return pileup_string
        elif self.type == 'indel':
            pileup_string = "Indel: {0} at: {1}\tRef: {2}\tCall: {3}\n".format(self.rname, self.pos, self.ref_base, self.con_base)
            return pileup_string       
    
    # Utility functions        
    def is_SNP(self):
        """Is this a SNP call?"""
        return True if (self.type == 'SNP') else False

    def is_indel(self):
        """Is this an indel call?"""    
        return True if (self.type == 'indel') else False

    def is_homozygous(self):
        """Is this a homozygous SNP?"""
        return True if (self.type == 'SNP') and (self.con_base in 'ACGT') else False

    def is_heterozygous(self):
        """is this a heterozygous SNP?"""
        return True if (self.type == 'SNP') and (self.con_base in 'MRWSYK') else False

        
        
if __name__ == '__main__':
    print "Hey, it's pileup_record.py!"
    
    in_handle = open(sys.argv[1],'r')
    for line in in_handle:
        print line
        p = Pileup_record(line)
        if p.is_SNP():
            print p
            print p.raw
        #print "SNP? {0} Homozygous? {1}".format(p.is_SNP(), p.is_homozygous())
    
    
"""
SAMtools pileup format:

From: http://sourceforge.net/apps/mediawiki/samtools/index.php?title=SAM_FAQ#I_do_not_understand_the_columns_in_the_pileup_output.

0. reference sequence name
1. reference coordinate
2. reference base, or `*' for an indel line
3. genotype where heterozygotes are encoded in the IUB code: M=A/C, R=A/G, W=A/T, S=C/G, Y=C/T and K=G/T; indels are indicated by, for example, */+A, -A/* or +CC/-C. There is no difference between */+A or +A/*.
4. Phred-scaled likelihood that the genotype is wrong, which is also called `consensus quality'.
5. Phred-scaled likelihood that the genotype is identical to the reference, which is also called `SNP quality'. Suppose the reference base is A and in alignment we see 17 G and 3 A. We will get a low consensus quality because it is difficult to distinguish an A/G heterozygote from a G/G homozygote. We will get a high SNP quality, though, because the evidence of a SNP is very strong.
6. root mean square (RMS) mapping quality
7. # reads covering the position
8. read bases at a SNP line (check the manual page for more information); the 1st indel allele otherwise
    At this column, a dot stands for a match to the reference base on the forward strand, a comma for a match on the reverse strand, ‘ACGTN’ for a mismatch on the forward strand and ‘acgtn’ for a mismatch on the reverse strand.
9. base quality at a SNP line; the 2nd indel allele otherwise
10. indel line only: # reads directly supporting the 1st indel allele
11. indel line only: # reads directly supporting the 2nd indel allele
12. indel line only: # reads supporting a third indel allele

If pileup is invoked without `-c', indel lines and columns between 3 and 7 inclusive will not be outputted. (col 2-6)

e.g.:

1	61	T	C	2	2	17	4	..C,	C>C.
1	85	T	T	45	0	19	6	...,,.+2AA	@5C+#/
1	85	*	*/+AA	20	20	19	6	*	+AA	5	1	0	0	0
1	194	A	A	30	0	37	1	.+1C	G
1	194	*	+C/+C	40	0	37	1	+C	*	1	0	0	0	0

"""    