#!/usr/bin/env python
# encoding: utf-8
"""
sam_to_oriented_fasta_new.py

Created by Rod Docking on 2010-09-01.
Copyright (c) 2010 Canada's Michael Smith Genome Sciences Centre. All rights reserved.
"""

import sys

from optparse import OptionParser
from common_utils import reverse_complement
from SAM_record import SAM_record

def quality_trim_seq(S, threshold):
    """Replace low-quality bases at the end of the read with X characters"""
    
    # Note - this assumes a specific quality-format for quality characters
    
    seq_list = list(S.seq)
    qual_list = S.qual_list()
    
    for i in range(len(qual_list)):
        if qual_list[i] >= threshold:
            break
        else:
            seq_list[i] = 'X'
    
    for i in range(len(qual_list)-1, -1, -1):
        if qual_list[i] >= threshold:
            break
        else:
            seq_list[i] = 'X'
    
    S.seq = ("").join(seq_list)
    
    return S

def parse_options():
    """Parse options provided on the command line"""
    
    usage = """usage: samtools view 10:500-2500 | %prog [options] -i200 1000:2000 reads_fwd.fa reads_rev.fa"""
    
    parser = OptionParser(usage)
    parser.add_option("-i", "--insert_size",
        action="store", type="int", dest="isize",
        default=200, help="mean fragment size")
    parser.add_option("-q", "--qual_trim",
        action="store", type="int", dest="qtrim",
        default=0, help="phred quality threshold for read trimming")
    
    (options, args) = parser.parse_args()
    
    return options, args


def main():
    """From SAM records in stdin, sort and orient reads for assembly"""
    
    # Parse options and infer read retrieval coordinates
    options, args = parse_options()
    
    start, stop = int(args[0].split('-')[0]), int(args[0].split('-')[1])
    region_one_list, region_two_list, region_three_list = [], [], []
    region_one_counter, region_two_counter, region_three_counter = 0, 0, 0
    
    fwd_file, rev_file = args[1], args[2]
    
    # sys.stderr.write("Region 1: {0}-{1}\n".format((start - int(options.isize*1.5)), start))
    # sys.stderr.write("Region 2: {0}-{1}\n".format(start, stop))
    # sys.stderr.write("Region 3: {0}-{1}\n".format(stop, (stop + int(options.isize*1.5))))
    
    # Read SAM input from stdin and sort reads
    found_good_seed = False
    for line in sys.stdin:
        
        S = SAM_record(line)
        
        if (options.qtrim > 0):
            S = quality_trim_seq(S, options.qtrim)
        
        if not S.edit_distance:
            S.edit_distance = 0
        
        # Split logic depending on alignment position
        # Region 1: only unmapped mates need to be kept for assembly
        if (S.pos < start) and (S.pos > (start - (options.isize*1.5))):
            region_one_counter += 1
            
            # Query unmapped, mate is on forward strand
            if S.q_unmapped() and (not S.mate_reverse_strand()):
                S.qname = "R1_{0}_".format(S.flag) + S.qname
                S.seq = reverse_complement(S.seq)
                region_one_list.append(S)
        
        # Region 2: all reads need to be kept for assembly
        elif ((S.pos >= start) and (S.pos <= stop)):
            
            # Find a good seed - mapped along it's entire length with no SNPs
            if not (found_good_seed):
                if "S" in S.cigar:
                    continue
                elif S.q_unmapped():
                    continue
                elif S.edit_distance > 0:
                    continue
                elif 'X' in S.seq:
                    continue
                elif 'N' in S.seq:
                    continue
                else:
                    found_good_seed = True
            
            region_two_counter += 1
            S.qname = "R2_{0}_".format(S.flag) + S.qname
            
            # Reverse-complement if...
            # Query mapped to reverse strand - NOTE - as below, there seem to be unmapped reads marked as mapped to the reverse strand
            if S.reverse_strand() and S.q_unmapped():
                S.seq = reverse_complement(S.seq)
            
            region_two_list.append(S)
        
        # Region 3: only unmapped mates need to be kept for assembly
        elif (S.pos > stop) and (S.pos < (stop + (options.isize*1.5))):
            region_three_counter += 1
            
            # Query unmapped, mate is on reverse strand
            if S.q_unmapped() and (S.mate_reverse_strand()):
                
                # Workaround - unmapped reads given 'reverse strand' flag??
                if S.reverse_strand():
                    S.seq = reverse_complement(S.seq)
                
                S.qname = "R3_{0}_".format(S.flag) + S.qname
                region_three_list.append(S)
    
    sys.stderr.write("{0:6}|{1:6}|{2:6}|".format(len(region_one_list),len(region_two_list),len(region_three_list)))
    
    # Write forward-orientation file
    fwd_handle = open(fwd_file, 'a')
    for S in (region_two_list + region_one_list + region_three_list):
        fwd_handle.write((">{0}\n{1}\n").format(S.qname, S.seq))
    fwd_handle.close()
    
    # Write reverse-orientation file
    region_one_list.reverse()
    region_two_list.reverse()
    region_three_list.reverse()
    
    found_good_seed = False
    rev_handle = open(rev_file, 'a')
    for S in region_two_list:
        if not (found_good_seed):
            if "S" in S.cigar:
                continue
            elif S.q_unmapped():
                continue
            elif 'X' in S.seq:
                continue
            elif 'N' in S.seq:
                continue
            else:
                found_good_seed = True
        
        rev_handle.write((">{0}\n{1}\n").format(S.qname, reverse_complement(S.seq)))
    
    for S in (region_one_list + region_three_list):
        rev_handle.write((">{0}\n{1}\n").format(S.qname, reverse_complement(S.seq)))
    rev_handle.close()

if __name__ == '__main__':
    main()

